/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2002 Kamil Skalski.
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: patrange.g,v 1.1 2002/08/16 12:17:30 malekith Exp $
 *
 * Patrange - check for pattern exhaustiveness and unused cases.
 *
 * The algorithm used here was developed by Kamil Skalski. I just recoded
 * it from OCaml to Gont. However the generate example code in old compiler
 * was wrong, for example for:
 *
 *	switch [x, y] {
 *	case [true, false]:
 *	case [false, true]:
 *	case [true, true]:
 *	}
 * 
 * it generates: [_, false]. However as I was unable to figure out how
 * to do it The Right Way, I use randomized pseudo-algorithm of my own,
 * to generate example.
 */

// TODO: fix generate example not to be random ;)

// someday-TODO: what to do:
// switch ex {
// case Foo: 
// 	// ...
// case Baar: 
//	// ...
// }
// where Baar is spelling error? The problem isn't as big in switch, but
// it is in with { ... } matching. -Wcapitalization has been now implemented,
// that warns about capitalized pattern variables. We might want 
// -Wstrict-capitalization, that warns about non-capitalize constructors etc.
// -Wcapitalization should only warn about potentially dangerous things.

// someday-TODO: unused clause checks for with-matching

open Ast;
open Util;
open Type;
open Patrange;

exception void No_example;

union range {
	void Omega;	// match-all, `_' or `x'
	bool BoolRange;
	<<range>List::t>List::t TupleRange;
	*[var_symbol, <range>List::t] UnionRange;
	expr ExprRange;
}

int expr_id;

string string_of_range(range r)
{
	string xalternatives(<range>List::t l)
	{
		let sl = List::qsort(String::cmp, 
				     List::map(string_of_range, l)) in
		return String::concat("|", sl);
	}
	
	switch r {
	case Omega:		return "_";
	case BoolRange[true]:	return "T";
	case BoolRange[false]:	return "F";
	case ExprRange[_]:	return "X<" + itoa(expr_id++) + ">";
		
	case TupleRange[lst]:
		return "[" + 
			String::concat(", ", List::map(xalternatives, lst)) + 
		       "]";
		       
	case UnionRange[s, lst]:
		return symbol_name(s) + "(" + xalternatives(lst) + ")";
	}
}

string alternatives(<range>List::t l)
{
	let sl = List::qsort(String::cmp, 
			     List::map(string_of_range, l)) in
	return String::concat(" | ", sl);
}
	

string string_of_example(range r)
{
	string just_one(<range>List::t lst)
	{
		switch lst {
		case List::Cons[x, List::Nil]:
			return string_of_example(x);
		case _:
			return ice("Patrange::string_of_example(): non-1");
		}
	}
	
	switch r {
	case Omega:		return "_";
	case BoolRange[true]:	return "true";
	case BoolRange[false]:	return "false";
	case ExprRange[e]:	return "some " + string_of_type(type_of(e));
		
	case TupleRange[lst]:
		return "[" + 
			String::concat(", ", List::map(just_one, lst)) + 
		       "]";
		       
	case UnionRange[s, lst]:
		return symbol_name(s) + "(" + just_one(lst) + ")";
	}
}

// check whatever big >= small
bool is_subrange_of(range big, range small)
{
	switch [big, small] {
	case [Omega, _]:
		return true;
		
	case [BoolRange[x], BoolRange[y]]:
		return x == y;
		
	case [TupleRange[l1], TupleRange[l2]]:
		void cross_check(<range>List::t b, <range>List::t s)
		{
			bool check_one(range ss)
			{
				let f = fun bool (range bb) 
						(is_subrange_of(bb, ss)) in
				return !List::exists(f, b);
			}

			if (List::exists(check_one, s))
				raise Exit;
		}

		try {
			List::iter2(cross_check, l1, l2);
			return true;
		} with {
		case Exit:
			return false;
		}
		
	case [UnionRange[s1, rr1], UnionRange[s2, r2]]:
		<range>List::t r1 = rr1;

		if (!same_symbols(s1, s2))
			return false;

		bool check_one(range s)
		{
			return !List::exists(fun bool (range b) 
							(is_subrange_of(b, s)),
					     r1);
		}

		return !List::exists(check_one, r2);
	
	case [_, _]:
		return false;
	}
}

// speed-TODO: rewrite
bool is_same_range(range r1, range r2)
{
	return string_of_range(r1) == string_of_range(r2);
}

// likewise
bool same_alternatives(<range>List::t l1, <range>List::t l2)
{
	if (List::length(l1) != List::length(l2))
		return false;
	
	let s1 = List::qsort(String::cmp, List::map(string_of_range, l1)) in
	let s2 = List::qsort(String::cmp, List::map(string_of_range, l2)) in
	return String::concat("@", s1) == String::concat("@", s2);
}

range xfold(range r)
{
	switch r {
	case TupleRange[lst]:
		bool isnt_ok(<range>List::t l)
		{
			switch l {
			case List::Cons[x, List::Nil]:
				switch xfold(x) {
				case Omega:
					return false;
				case _:
					return true;
				}
			case _:
				return true;
			}
		}

		if (List::exists(isnt_ok, lst))
			return r;
		else
			return Omega;

	case UnionRange[s, List::Cons[x, List::Nil]]:
		return UnionRange[s, List::cons1(xfold(x))];

	case _:
		return r;
	}
}

range fold(range r)
{
//dbg("(fold: " + string_of_range(r) + " -> ");
	r = xfold(r);
//dbg(string_of_range(r) + ")\n");
	return r;
}

range range_of_pattern(pattern p)
{
	range r;

	switch p.a {
	case VoidPattern:
		return Omega;
		
	case VarPattern[v]:
		return Omega;
	
	case ExprPattern[e]:
		switch e.a.raw {
		case BoolConst[x]:
			return BoolRange[x];
		case _:
			return ExprRange[e];
		}

		
	case TuplePattern[lst]:
		<range>List::t f(pattern pp)
		{
			return List::cons1(range_of_pattern(pp));
		}
		
		r = TupleRange[List::map(f, lst)];
	
	case UnionPattern[np]:
		let x = range_of_pattern(np.np_pattern) in
		r = UnionRange[get_binding(np.np_symbol), List::cons1(x)];
	}

	return fold(r);
}


<var_symbol>List::t union_fields(var_symbol one_field)
{
	var_symbol get_fld_name(<var_symbol>field f)
	{
		return get_binding(f.fld_symbol);
	}
	
	switch (union_of(one_field)).stype {
	case Union[x]:
		return List::map(get_fld_name, x);
	case _:
		return ice("Patrange::union_fields(): " 
			   + "bad parent union");
	}
}


<range>List::t union_check(<range>List::t lst)
{
	<var_symbol>List::t fields;
	
//dbg("(union_check: " + alternatives(lst) + "\n");

	switch lst {
	case List::Cons[UnionRange[s, _], _]:
		fields = union_fields(s);
		
	case _:
		return lst;
	}

	if (List::length(lst) != List::length(fields)) {
//dbg("union_check: bad ret)\n");
		return lst;
	}

	bool isnt_in_lst(var_symbol s)
	{
		bool is_ok(range r)
		{
			switch r {
			case UnionRange[ss, List::Cons[Omega, List::Nil]]:
				return same_symbols(s, ss);
				
			case UnionRange[_, _]:
				return false;
			
			case _:
				ice("Patrange::union_check(): " + 
					   "non-UnionRange");
				return false;
			}
		}

		return !List::exists(is_ok, lst);
	}

	if (!List::exists(isnt_in_lst, fields))
		lst = List::cons1(Omega);
//dbg("union_check_ret: " + alternatives(lst) + ")\n");
	return lst;
}

<range>List::t merge_in(<range>List::t lst, range new)
{
	new = fold(new);

	if (List::exists(fun bool (range x) (is_subrange_of(x, new)), lst)) {
//dbg("warn: " + string_of_range(new) + " to " + alternatives(lst) + "\n");
		warn("unused pattern");
		return lst;
	}
	
//dbg("(merge_in: " + string_of_range(new) + " to " + alternatives(lst) + "\n");
	<range>option find_place_in(range old)
	{
		switch [old, new] {
		case [_, Omega]:
			return Some[Omega];
		
		case [UnionRange[s1, l1], UnionRange[s2, l2]]:
			if (!same_symbols(s1, s2))
				return None;
			
			let n = List::fold_left(merge_in, l1, l2) in
			return Some[UnionRange[s1, n]];
	
		case [BoolRange[true], BoolRange[false]]:
			return Some[Omega];
			
		case [BoolRange[false], BoolRange[true]]:
			return Some[Omega];
		
		case [TupleRange[l1], TupleRange[l2]]:
			<<range>List::t>List::t ret = List::Nil;
			int pos = 0, diff_at = -1;

			void chk(<range>List::t x, <range>List::t y)
			{
				if (!same_alternatives(x, y)) {
					if (diff_at != -1)
						raise Exit;
					else
						diff_at = pos;
				}

				pos++;
			}

			void merge_em(<range>List::t x, <range>List::t y)
			{
				if (pos++ == diff_at)
					x = List::fold_left(merge_in, x, y);
				
				ret = List::Cons[x, ret];
			}


			try {
				List::iter2(chk, l1, l2);
				if (diff_at == -1)
					ice("Patrange::merge_in(): the same");
				pos = 0;
				List::iter2(merge_em, l1, l2);
				return Some[fold(TupleRange[List::rev(ret)])];
			} with {
			case Exit:
				return None;
			}
			
		case [_, _]:
			return None;
		}
	}

	<range>option merged = None;
	
	<range>List::t iter(<range>List::t a)
	{
		switch a {
		case List::Nil:
			return List::cons1(new);
			
		case List::Cons[x, xs]:
			switch find_place_in(x) {
			case Some[y]:
				merged = Some[y];
				return xs;
			case None:
				if (is_subrange_of(new, x))
					return iter(xs);
				else
					return List::Cons[x, iter(xs)];
			}
		}
	}

	lst = iter(lst);

	switch merged {
	case Some[x]:
		new = x;
		bool aint_subrange(range r)
		{
			return !is_subrange_of(new, r);
		}
	
		lst = merge_in(List::filter(aint_subrange, lst), new);
		
	case None:
		skip;
	}
	
	lst = union_check(lst);
//dbg("merge_in_ret: " + alternatives(lst) + ")\n");
	return lst;
}

void check_exhaustiveness(<pattern>List::t pats)
{
	<range>List::t ctx = List::Nil;

	if (!flags.wpattern)
		return;
	
	void check_one(pattern p)
	{
		push_location(p.loc);
		let r = range_of_pattern(p) in
		ctx = merge_in(ctx, r);
		pop_location();
	}

	List::iter(check_one, pats);

	switch ctx {
	case List::Cons[Omega, List::Nil]:
		return;		// ok
	case _:
		warn("this pattern matching is not exhaustive");

		try {
			string ex = string_of_example(generate_example(ctx));
			warn("example of value that isn't matched: " + ex);
		} with {
		case No_example:
			warn("cannot generate example in resoanable time, " +
			     "compiler's author is lame");
		}
	}
}

// check whether pat could match example
bool could_match_1(range pat, range example)
{
	switch [pat, example] {
	case [Omega, _]:
		return true;
		
	case [_, Omega]:
		return true;
		
	case [BoolRange[x], BoolRange[y]]:
		return x == y;
		
	case [TupleRange[l1], TupleRange[l2]]:
		void cross_check(<range>List::t pat_l, <range>List::t example_l)
		{
			range examp;
			
			switch example_l {
			case List::Cons[x, List::Nil]:
				examp = x;
			case _:
				ice("Patrange::could_match_1(): non-1");
				return;
			}
			
			if (!List::exists(fun bool (range pa) 
						(could_match_1(pa, examp)), 
					  pat_l))
				raise Exit;
		}

		try {
			List::iter2(cross_check, l1, l2);
			return true;
		} with {
		case Exit:
			return false;
		}
		
	case [UnionRange[s1, pat_l], UnionRange[s2, ex_l]]:
		if (!same_symbols(s1, s2))
			return false;

		switch ex_l {
		case List::Cons[exx, List::Nil]:
			range ex = exx;
			return List::exists(fun bool (range p) 
							(could_match_1(p, ex)),
					    pat_l);

		case _:
			ice("Patrange::could_match_1(): non-1?");
			return false;
		}

	case [_, _]:
		return false;
	}
}

bool could_match(<range>List::t pat_l, range ex)
{
	return List::exists(fun bool (range p) (could_match_1(p, ex)), pat_l);
}

int rand_seed;

int rand()
{
	// but these are taken from Bruce Schneier: Applied Cryptography
	rand_seed = (rand_seed * 4096 + 150889) % 714024;

	return rand_seed;
}

'a pick_random_from_list(<'a>List::t lst)
{
	int n = List::length(lst);

	if (n == 0)
		ice("Patrange::pick_random_from_list(): n = 0");

	n = rand() % n;

	while (n-- > 0)
		switch lst {
		case List::Cons[_, tl]:
			lst = tl;
		case List::Nil:
			ice("Patrange::pick_random_from_list(): impossible");
		}
		
	switch lst {
	case List::Cons[hd, _]:
		return hd;
	case List::Nil:
		return ice("Patrange::pick_random_from_list(): impossible 2");
	}
}

range union_example(var_symbol us, <range>List::t ranges)
{
	<var_symbol, <range>List::t>Ptrhash::t matched = 
					Ptrhash::create(13);
	
	void add(range r)
	{
		switch r {
		case UnionRange[s, l]:
			Ptrhash::add(matched, s, l);
		case _:
			ice("Patrange::union_example(): non-union");
		}
	}

	List::iter(add, ranges);

	<range>List::t ret = List::Nil;

	void check(var_symbol s)
	{
		if (Ptrhash::mem(matched, s))
			let rs = Ptrhash::find(matched, s) in
			try {
				let ex = List::cons1(generate_example(rs)) in
				ret = List::Cons[UnionRange[s, ex], ret];
			} with {
			case No_example:
				skip;
			}
		else
			let p = UnionRange[s, List::cons1(Omega)] in
			ret = List::Cons[p, ret];
	}

	List::iter(check, union_fields(us));

	return pick_random_from_list(ret);
}

range tuple_example(<<range>List::t>List::t lst, <range>List::t rest)
{
	<range>List::t examples = List::Nil;

	<<range>List::t>List::t h1(<<range>List::t>List::t so_far,
				   <<range>List::t>List::t out_of)
	{
		switch out_of {
		case List::Cons[hd, tl]:
			let rnd = pick_random_from_list(hd) in
			let rec = h1(List::Cons[hd, so_far], tl) in
			let ret = List::Cons[List::cons1(rnd), rec] in
			try {
				let ex = List::cons1(generate_example(hd)) in
				let exa = TupleRange[
					List::concat(List::rev(so_far),
					List::Cons[ex, rec])] in
				examples = List::Cons[exa, examples];
				return ret;
			} with {
			case No_example:
				return ret;
			}

		case List::Nil:
			return List::Nil;
		}
	}

	bool try_it_out(range ex)
	{
//	dbg("ex: " + string_of_range(ex) + "\n");
		return !could_match(rest, ex);
	}
	
	// try few times
	for (int i = 0; i < 32; i++) {
//dbg("i="+itoa(i)+"\n");
		h1(List::Nil, lst);
		try {
			return List::find(try_it_out, examples);
		} with {
		case Not_found:
			skip;
		}
		examples = List::Nil;
	}

	raise No_example;
}

// Generate random example. We will stick to it until we'll get something
// more reasonable.
range generate_example(<range>List::t ranges)
{
	void check_nil(<range>List::t nil)
	{
		switch nil {
		case List::Nil:
			skip;
		case _:
			ice("Patrange::generate_example(): should be 1-long: " +
				alternatives(ranges));
		}
	}
	
	switch ranges {
	case List::Cons[BoolRange[x], nil]:
		// there cannot be BoolRange[!x] in list
		check_nil(nil);
		return BoolRange[!x];
		
	case List::Cons[Omega, nil]:
		check_nil(nil);
		raise No_example;
	
	case List::Cons[ExprRange[e], lst]:
		bool aint_expr(range r)
		{
			switch r {
			case ExprRange[_]:	return false;
			case _:			return true;
			}
		}
		
		check_nil(List::filter(aint_expr, lst));

		return ExprRange[e];

	case List::Cons[UnionRange[us, _], _]:
		return union_example(us, ranges);
	
	case List::Cons[TupleRange[lst], rest]:
		return tuple_example(lst, rest);
	
	case List::Nil:
		return ice("Patrange::generate_example(): nil");
	}
}

